/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.jdbc.dao;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.registry.RegistryConstants;
import org.wso2.registry.Resource;
import org.wso2.registry.RegistryException;
import org.wso2.registry.User;
import org.wso2.registry.utils.AuthorizationUtil;
import org.wso2.registry.jdbc.DatabaseConstants;
import org.wso2.usermanager.Realm;

import java.sql.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

public class VersionedResourceDAO {

    private static final Log log = LogFactory.getLog(VersionedResourceDAO.class);

    /**
     * Returns the given version of the artifact. If artifact is a collection all child paths will
     * be added as versioned paths in the form /a/b?v=12
     *
     * @param path          : path of the artifact
     * @param versionNumber : Version number
     * @param conn          : Connection to DB
     * @return Resource for the given version number
     * @throws SQLException : If something went wrong
     */
    public Resource get(String path, long versionNumber, Connection conn) throws SQLException {

        Resource resource = null;

        ResultSet result;

        // get the resource data from the resource table

        String sql = "SELECT * FROM ARTIFACTS A WHERE A.PATH=?";
        PreparedStatement statement = conn.prepareStatement(sql);
        statement.setString(1, path);

        result = statement.executeQuery();

        if (result.next()) {
            resource = new Resource();

            resource.setId(result.getLong(DatabaseConstants.AID_FIELD));
            resource.setPath(path + "?v=" + versionNumber);
            resource.setParentPath(getParentPath(resource.getId(), versionNumber, conn));
            resource.setDirectory(result.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
            resource.setMediaType(result.getString(DatabaseConstants.MEDIA_TYPE_FIELD));
            resource.setState(result.getInt(DatabaseConstants.STATE_FIELD));
            resource.setAuthorUserName(result.getString(DatabaseConstants.AUTHOR_FIELD));
            resource.setCreatedTime(result.getTimestamp(DatabaseConstants.CREATED_TIME_FIELD));
            resource.setDescription(result.getString(DatabaseConstants.DESCRIPTION_FIELD));
        }

        if (resource == null) {
            return null;
        }

        String sqlVersionedFields = "SELECT * FROM VERSIONS WHERE AID=? AND VN=?";
        PreparedStatement versionQuery = conn.prepareStatement(sqlVersionedFields);
        versionQuery.setLong(1, resource.getId());
        versionQuery.setLong(2, versionNumber);

        ResultSet versionResults = versionQuery.executeQuery();

        if (versionResults.next()) {
            resource.setLastUpdaterUserName(
                    versionResults.getString(DatabaseConstants.VERSION_AUTHOR_FIELD));
            resource.setLastModified(
                    versionResults.getTimestamp(DatabaseConstants.VERSION_UPDATED_TIME));
            resource.setContent(versionResults.getBytes(DatabaseConstants.VERSION_CONTENT_FIELD));
        }

        // get resource children

        if (resource.isDirectory()) {

            String[] children = getChildPaths(resource.getId(), versionNumber, conn);
            resource.setContent(children);
        }

        // get resource properties

        String propSQL = "SELECT P.PKEY, P.PVALUE FROM PROPERTIES P WHERE P.AID=?";
        PreparedStatement s3 = conn.prepareStatement(propSQL);
        s3.setLong(1, resource.getId());
        ResultSet propResults = s3.executeQuery();

        Properties props = new Properties();
        while (propResults.next()) {
            String key = propResults.getString(DatabaseConstants.PROPERTY_KEY_FIELD);
            String value = propResults.getString(DatabaseConstants.PROPERTY_VALUE_FIELD);
            props.put(key, value);
        }
        resource.setProperties(props);

        return resource;
    }

    public Resource getResourceByID(long artifactID, long versionNumber, Connection conn)
            throws SQLException {

        Resource resource = null;

        ResultSet result;

        // get the resource data from the resource table

        String sql = "SELECT * FROM ARTIFACTS A WHERE A.AID=?";
        PreparedStatement statement = conn.prepareStatement(sql);
        statement.setLong(1, artifactID);

        result = statement.executeQuery();

        if (result.next()) {
            resource = new Resource();

            resource.setId(result.getLong(DatabaseConstants.AID_FIELD));
            resource.setPath(result.getString(DatabaseConstants.PATH_FIELD));
            resource.setDirectory(result.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
            resource.setMediaType(result.getString(DatabaseConstants.MEDIA_TYPE_FIELD));
            resource.setState(result.getInt(DatabaseConstants.STATE_FIELD));
            resource.setAuthorUserName(result.getString(DatabaseConstants.AUTHOR_FIELD));
            resource.setCreatedTime(result.getTimestamp(DatabaseConstants.CREATED_TIME_FIELD));
            resource.setDescription(result.getString(DatabaseConstants.DESCRIPTION_FIELD));
        }

        if (resource == null) {
            return null;
        }

        if (versionNumber == -1) {
            versionNumber = getLatestVerisonNumber(artifactID, conn);
        }
        String sqlVersionedFields = "SELECT * FROM VERSIONS WHERE AID=? AND VN=?";
        PreparedStatement versionQuery = conn.prepareStatement(sqlVersionedFields);
        versionQuery.setLong(1, resource.getId());
        versionQuery.setLong(2, versionNumber);

        ResultSet versionResults = versionQuery.executeQuery();

        if (versionResults.next()) {
            resource.setLastUpdaterUserName(
                    versionResults.getString(DatabaseConstants.VERSION_AUTHOR_FIELD));
            resource.setLastModified(
                    versionResults.getTimestamp(DatabaseConstants.VERSION_UPDATED_TIME));
            resource.setContent(versionResults.getBytes(DatabaseConstants.VERSION_CONTENT_FIELD));
        }

        // get resource children

        if (resource.isDirectory()) {

            String[] children = getChildPaths(resource.getId(), versionNumber, conn);
            resource.setContent(children);
        }

        // get resource properties

        String propSQL = "SELECT P.PKEY, P.PVALUE FROM PROPERTIES P WHERE P.AID=?";
        PreparedStatement s3 = conn.prepareStatement(propSQL);
        s3.setLong(1, resource.getId());
        ResultSet propResults = s3.executeQuery();

        Properties props = new Properties();
        while (propResults.next()) {
            String key = propResults.getString(DatabaseConstants.PROPERTY_KEY_FIELD);
            String value = propResults.getString(DatabaseConstants.PROPERTY_VALUE_FIELD);
            props.put(key, value);
        }
        resource.setProperties(props);

        return resource;
    }

    /**
     * Returns the latest (current) version of an artifact. If artifact is a collection, all child
     * paths will be added as latest paths in the form /a/b
     *
     * @param path : path of the artifact
     * @param conn :Connection to DB
     * @return : Result artifact
     * @throws SQLException : if something went wrong
     */
    public Resource getLatestVersion(String path, Connection conn) throws RegistryException , SQLException {
        /*
        * when someone give path like "root/wso2/wsas/r1" , then need to figure out the latest
        * version of the root and from the need to figure out the version of wso2 and so so
        * so when list outing the nodes need to give them correctly.
        * */

        Resource resource = null;

        ResultSet result;

        // get the resource data from the resource table

        String sql = "SELECT * FROM ARTIFACTS A WHERE A.PATH=?";
        PreparedStatement statement = conn.prepareStatement(sql);
        statement.setString(1, path);

        result = statement.executeQuery();

        if (result.next()) {
            resource = new Resource();

            resource.setId(result.getLong(DatabaseConstants.AID_FIELD));
            resource.setPath(result.getString(DatabaseConstants.PATH_FIELD));
            resource.setParentPath(getParentPath(resource.getPath()));
            resource.setDirectory(result.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
            resource.setMediaType(result.getString(DatabaseConstants.MEDIA_TYPE_FIELD));
            resource.setState(result.getInt(DatabaseConstants.STATE_FIELD));
            resource.setAuthorUserName(result.getString(DatabaseConstants.AUTHOR_FIELD));
            resource.setCreatedTime(result.getTimestamp(DatabaseConstants.CREATED_TIME_FIELD));
            resource.setDescription(result.getString(DatabaseConstants.DESCRIPTION_FIELD));
        }

        if (resource == null) {
            return null;
        }

//        long latestVersionNumber = getLatestVerisonNumber(resource.getId(), conn);
        long latestVersionNumber = getCurrentVerisonNumber(resource.getPath(), conn);


        String sqlVersionedFields = "SELECT * FROM VERSIONS WHERE AID=? AND VN=?";
        PreparedStatement versionQuery = conn.prepareStatement(sqlVersionedFields);
        versionQuery.setLong(1, resource.getId());
        versionQuery.setLong(2, latestVersionNumber);

        ResultSet versionResults = versionQuery.executeQuery();

        if (versionResults.next()) {
            resource.setLastUpdaterUserName(
                    versionResults.getString(DatabaseConstants.VERSION_AUTHOR_FIELD));
            resource.setLastModified(
                    versionResults.getTimestamp(DatabaseConstants.VERSION_UPDATED_TIME));
            resource.setContent(versionResults.getBytes(DatabaseConstants.VERSION_CONTENT_FIELD));
        }

        // get resource children
        if (resource.isDirectory()) {
            String[] children = getLatestChildPaths(resource.getId(),
                                                    getCurrentVerisonNumber(path, conn), conn);
            resource.setContent(children);
        }

        // get resource properties
        String propSQL = "SELECT P.PKEY, P.PVALUE FROM PROPERTIES P WHERE P.AID=?";
        PreparedStatement s3 = conn.prepareStatement(propSQL);
        s3.setLong(1, resource.getId());
        ResultSet propResults = s3.executeQuery();

        Properties props = new Properties();
        while (propResults.next()) {
            String key = propResults.getString(DatabaseConstants.PROPERTY_KEY_FIELD);
            String value = propResults.getString(DatabaseConstants.PROPERTY_VALUE_FIELD);
            props.put(key, value);
        }
        resource.setProperties(props);
        return resource;
    }

    /**
     * To get the current version for a give resource , current version may not be the latest
     * version in all the cases. If someone restore a resource then the child nodes will remain same
     * But the parent version will be changed. In that case the parent [foo/bar] version will be the
     * latest version but the child node version [foo/bar/v1] will be an old version . So when some
     * one ask for the resource value foo/bar/v1 then we need to calculate them correctly and give
     * that.
     *
     * @param path : Resource ID
     * @param conn : Connection to DB
     * @return : verion number of the resource user requested
     */
    private long getCurrentVerisonNumber(String path, Connection conn) throws RegistryException {
        if (path != null) {
            try {
                String[] parts = path.split(RegistryConstants.PATH_SEPARATOR);
                if(parts.length > 1 ) {
                    String currentPath = RegistryConstants.ROOT_PATH + parts[1];
                    String parentPath  = RegistryConstants.ROOT_PATH;
                    long parentVersion = getLatestVerisonNumber(
                            getResourceID(RegistryConstants.ROOT_PATH, conn), conn);
                    long childId = 0;
                    for (int i = 2; i < parts.length; i++) {
                        if (!resourceActive(parentPath, conn)){
                            throw new RegistryException("Resource is not active");
                        }
                        long parentId = getResourceID(parentPath, conn);
                        childId = getResourceID(currentPath, conn);
                        parentVersion = getChildVersion(conn, parentId, parentVersion, childId);
                        parentPath = currentPath;
                        currentPath = currentPath + RegistryConstants.PATH_SEPARATOR + parts[i];
                    }
                    long id = getResourceID(currentPath,conn);
                    return getChildVersion(conn, childId, parentVersion, id);

                } else {
                    return getLatestVerisonNumber(getResourceID(path, conn), conn);
                }
//                int index = path.lastIndexOf(RegistryConstants.PATH_SEPARATOR);
//                if (index > 0) {
//                    String parentPath = path.substring(0, index);
//                    long parentId = getResourceID(parentPath, conn);
//                    long latestVersion = getLatestVerisonNumber(parentId, conn);
//                    long childID = getResourceID(path, conn);
//                    String SQL = "SELECT DVN FROM DEPENDENCY WHERE AID=? AND VN=? AND DAID=?";
//                    PreparedStatement s = conn.prepareStatement(SQL);
//                    s.setLong(1, parentId);
//                    s.setLong(2, latestVersion);
//                    s.setLong(3, childID);
//                    ResultSet result = s.executeQuery();
//                    if (result.next()) {
//                        return result.getLong(1);
//                    }
//                } else {
//                    return getLatestVerisonNumber(getResourceID(path, conn), conn);
//                }
            } catch (SQLException e) {
                log.info("Something went wrong while calculating the current " +
                         "version for the path : " + path + " and the error us " + e);
                return -1;
            }
        }
        return -1;
    }

    private long getChildVersion(Connection conn,
                                long parentId,
                                long parentVersion,
                                long childId) throws SQLException {
        String SQL = "SELECT DVN FROM DEPENDENCY WHERE AID=? AND VN=? AND DAID=?";
        PreparedStatement selectps = conn.prepareStatement(SQL);
        selectps.setLong(1, parentId);
        selectps.setLong(2, parentVersion);
        selectps.setLong(3, childId);
        ResultSet childResult = selectps.executeQuery();
        while (childResult.next()) {
            return childResult.getLong("DVN");
        }
        return -1;
    }

    public String[] getChildPaths(long artifcatID, long versionNumber,
                                  Connection connection) throws SQLException {

        String selectSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
        PreparedStatement s = connection.prepareStatement(selectSql);
        s.setLong(1, artifcatID);
        s.setLong(2, versionNumber);
        ResultSet result = s.executeQuery();

        ArrayList dependencyPathList = new ArrayList();

        //to see whether the patent node is there in the table , if not need to add that
        while (result.next()) {
            long daid = result.getLong("DAID");
            long childVersion = result.getLong("DVN");
            String SQL = "SELECT PATH FROM ARTIFACTS WHERE AID=?";
            PreparedStatement ps = connection.prepareStatement(SQL);
            ps.setLong(1, daid);
            ResultSet pathResult = ps.executeQuery();
            if (pathResult.next()) {
                String path = pathResult.getString("PATH") + "?v=" + childVersion;
                dependencyPathList.add(path);
            }
        }
        return (String[])(dependencyPathList.toArray(new String[dependencyPathList.size()]));
    }

    public String getParentPath(long resourceID, long versionNumber,
                                Connection connection) throws SQLException {

        if (versionNumber == -1) {
            versionNumber = getLatestVerisonNumber(resourceID, connection);
        }
        if (versionNumber == -1) {
            return null;
        }
        String selectSql = "SELECT AID , VN FROM DEPENDENCY WHERE DAID=? AND DVN=?";
        PreparedStatement s = connection.prepareStatement(selectSql);
        s.setLong(1, resourceID);
        s.setLong(2, versionNumber);
        ResultSet result = s.executeQuery();

        String parentPath = "";

        //to see whether the patent node is there in the table , if not need to add that
        if (result.next()) {
            long aid = result.getLong("AID");
            long parentVersionNumber = result.getLong("VN");
            String SQL = "SELECT PATH FROM ARTIFACTS WHERE AID=?";
            PreparedStatement ps = connection.prepareStatement(SQL);
            ps.setLong(1, aid);
            ResultSet pathResult = ps.executeQuery();
            if (pathResult.next()) {
                String path = pathResult.getString("PATH");
                parentPath = path + "?v=" + parentVersionNumber;
            }
        }
        return parentPath;
    }

    public String[] getLatestChildPaths(long resourceID, long versionNumber,
                                        Connection connection) throws SQLException {

        String selectSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
        PreparedStatement s = connection.prepareStatement(selectSql);
        s.setLong(1, resourceID);
        s.setLong(2, versionNumber);
        ResultSet result = s.executeQuery();

        ArrayList dependencyPathList = new ArrayList();

        //to see whether the patent node is there in the table , if not need to add that
        while (result.next()) {
            long daid = result.getLong("DAID");
            String SQL = "SELECT PATH FROM ARTIFACTS WHERE AID=?";
            PreparedStatement ps = connection.prepareStatement(SQL);
            ps.setLong(1, daid);
            ResultSet pathResult = ps.executeQuery();
            if (pathResult.next()) {
                String path = pathResult.getString("PATH");
                dependencyPathList.add(path);
            }
        }
        return (String[])(dependencyPathList.toArray(new String[dependencyPathList.size()]));
    }

    public List getChildren(long resourceID, long versionNumber,
                            Connection connection) throws SQLException {

        if (versionNumber == -1) {
            versionNumber = getLatestVerisonNumber(resourceID, connection);
        }

        String selectSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
        PreparedStatement s = connection.prepareStatement(selectSql);
        s.setLong(1, resourceID);
        s.setLong(2, versionNumber);
        ResultSet result = s.executeQuery();

        ArrayList childList = new ArrayList();

        //to see whether the patent node is there in the table , if not need to add that
        while (result.next()) {

            long daid = result.getLong("DAID");
            String SQL = "SELECT AID, PATH, DIRECTORY FROM ARTIFACTS WHERE AID=?";
            PreparedStatement ps = connection.prepareStatement(SQL);
            ps.setLong(1, daid);

            ResultSet pathResult = ps.executeQuery();
            if (pathResult.next()) {
                Resource child = new Resource();
                child.setId(pathResult.getLong(DatabaseConstants.AID_FIELD));
                child.setPath(pathResult.getString(DatabaseConstants.PATH_FIELD));
                child.setDirectory(pathResult.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
                childList.add(child);
            }
        }
        return childList;
    }

    /**
     * Adds the given resource to the database.
     *
     * @param resource New resource
     * @param conn     : Connection to Db
     * @throws java.sql.SQLException : if something went wrong
     */
    public void add(Resource resource, Connection conn) throws SQLException {

        // first add the resource data into the artifacts table

        String sql =
                "INSERT INTO ARTIFACTS (PATH, MEDIA_TYPE, DIRECTORY, STATE, CREATED_TIME, AUTHOR, " +
                        "DESCRIPTION) VALUES (?,?,?,?,?,?,?)";

        long now = System.currentTimeMillis();

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, resource.getPath());
        s.setString(2, resource.getMediaType());
        s.setInt(3, resource.isDirectory() ? 1 : 0);
        s.setInt(4, RegistryConstants.ACTIVE_STATE);
        s.setTimestamp(5, new Timestamp(now));
        s.setString(6, resource.getAuthorUserName());
        s.setString(7, resource.getDescription());

        s.executeUpdate();


        long newArtifactID = this.getResourceID(resource.getPath(), conn);

        // now add resource properties to the properties table
        addProperties(newArtifactID, resource.getProperties(), conn);
    }

    public void update(String path, Resource resource, Connection conn) throws SQLException {

        // first update the artifacts table

        String sql =
                "UPDATE ARTIFACTS SET MEDIA_TYPE=?, DIRECTORY=?, STATE=?, AUTHOR=?, DESCRIPTION=? WHERE PATH=?";

        PreparedStatement s = conn.prepareStatement(sql);

        s.setString(1, resource.getMediaType());
        s.setInt(2, resource.isDirectory() ? 1 : 0);
        s.setInt(3, resource.getState());
        s.setString(4, resource.getAuthorUserName());
        s.setString(5, resource.getDescription());

        s.setString(6, path);

        s.executeUpdate();

        // remove existing properties for the updated artifact
        removeProperties(resource.getId(), conn);

        // now add the updated properties to the properties table
        addProperties(resource.getId(), resource.getProperties(), conn);
    }

    public void markDeleted(String path, Connection conn) throws SQLException {

        String sql = "UPDATE ARTIFACTS SET STATE=? WHERE PATH=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setInt(1, RegistryConstants.DELETED_STATE);
        s.setString(2, path);
        s.executeUpdate();
    }

     public void markActivated(long id, Connection conn) throws SQLException {

        String sql = "UPDATE ARTIFACTS SET STATE=? WHERE AID=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setInt(1, RegistryConstants.ACTIVE_STATE);
        s.setLong(2, id);
        s.executeUpdate();
    }


    public void addProperties(long resourceID, Properties props, Connection conn)
            throws SQLException {

        String propSQL = "INSERT INTO PROPERTIES (AID, PKEY, PVALUE) VALUES (?,?,?)";
        PreparedStatement s2 = conn.prepareStatement(propSQL);

        if (props != null) {

            Iterator i = props.keySet().iterator();
            while (i.hasNext()) {
                String key = (String)i.next();
                String value = (String)props.get(key);

                s2.setLong(1, resourceID);
                s2.setString(2, key);
                s2.setString(3, value);

                s2.executeUpdate();
                s2.clearParameters();
            }
        }
    }

    public void removeProperties(long resourceID, Connection conn) throws SQLException {

        String removePropsSQL = "DELETE FROM PROPERTIES WHERE AID=?";
        PreparedStatement s2 = conn.prepareStatement(removePropsSQL);
        s2.setLong(1, resourceID);
        s2.executeUpdate();
    }

    public boolean resourceExist(String path, Connection conn) throws SQLException {

        String sql = "SELECT COUNT(A.AID) FROM ARTIFACTS A WHERE A.PATH=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, path);

        ResultSet result = s.executeQuery();
        long count = 0;
        if (result.next()) {
            count = result.getLong(1);
        }

        return count == 1;
    }

     public boolean resourceActive(String path, Connection conn) throws SQLException {

        String sql = "SELECT STATE FROM ARTIFACTS A WHERE A.PATH=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, path);

        ResultSet result = s.executeQuery();
        long count = 0;
        if (result.next()) {
            count = result.getLong(1);
        }
        if (RegistryConstants.ACTIVE_STATE == count) {
            return  true;
        } else {
            return false;
        }
    }

    public long getResourceID(String path, Connection conn) throws SQLException {

        String sql = "SELECT A.AID FROM ARTIFACTS A WHERE A.PATH=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, path);

        ResultSet result = s.executeQuery();
        if (result.next()) {
            return result.getLong(DatabaseConstants.AID_FIELD);
        }

        return -1;
    }

    public String getResourcePath(long id, Connection conn) throws SQLException {

        String sql = "SELECT A.PATH FROM ARTIFACTS A WHERE A.AID=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setLong(1, id);

        ResultSet result = s.executeQuery();
        if (result.next()) {
            return result.getString(DatabaseConstants.PATH_FIELD);
        }
        return null;
    }

    /**
     * This method will move a resource from oldpath to newpath , the internal implementation is such
     * that frit the release will be deleted from oldPath parent and will be added to the newptah
     * so if the oldpath is "foo/bar/r1" then the r1 will be removed from foo/bar. However later
     * we can restore and get everything working fine
     * @param oldPath : Current path of the resource
     * @param newPath : Where to move the resource
     * @param conn : Connection to DB
     * @throws org.wso2.registry.RegistryException : If something went wrong
     * @param userId : current user
     * @param realm : Realm
     */
    public void renameResource(String oldPath,
                               String newPath,
                               Connection conn ,
                               String userId,
                               Realm realm) throws RegistryException {
        try {
            Resource resource = getLatestVersion(oldPath, conn);
            String resourcePath = resource.getPath();
            long resourceID = resource.getId();
             //to see whether the patent node is there in the table , if not need to add that
            createParentCollections(newPath, conn, userId, realm);
            // if the resource is a directory then need to populate the dependecy table with
            // the new resource id , while keeping the old one as it is
            if (resource.isDirectory()) {
                throw new RegistryException("Resource rename does not support for collections");
            } else {
                // Then its just need to add the resource and no need to update any of the
                // dependecy tables
                 resource.setPath(newPath);
                if (resource.getAuthorUserName() == null) {
                    resource.setAuthorUserName(userId);
                }
                resource.setLastUpdaterUserName(userId);
                add(resource, conn);
                AuthorizationUtil.setDefaultAuthorizations(realm, newPath, userId);
                long newResourceID = getResourceID(newPath, conn);
                resource.setId(newResourceID);
                addResourceVersion(resource, conn);
            }
            resource.setPath(resourcePath);
            resource.setId(resourceID);
            markDeleted(resourcePath, conn);
            deleteResource(resource, conn);
            //prepareToDelete(resource.getId(), conn);
            //resourceDAO.delete(resource.getResourcePath(), conn);
        } catch (SQLException e) {
            throw new RegistryException(e.getMessage());
        }
    }

    public long addResourceVersion(Resource resource, Connection connection) throws SQLException , RegistryException {

        if (resource.getLastModified() == null) {
            resource.setLastModified(new java.util.Date());
        }

        long versionNumber = insertToVersionTable(resource, connection);

        //updating max version table
        updateDependency(resource, versionNumber, false, connection);
        return versionNumber;
    }

    private long insertToVersionTable(Resource resource, Connection connection)
            throws SQLException {

        String sql =
                "INSERT INTO VERSIONS (AID, VN ,CONTENT , AUTHOR, UPDATED_TIME) VALUES (?,?,?,?,?)";
        PreparedStatement statement = connection.prepareStatement(sql);
        statement.setLong(1, resource.getId());

        // Need to get the old version and increment that
        // Select VID from latest table ,
        // if it is there then find from that  , if it is not there then we need ti start from 0
        long versionNumber = getLatestVerisonNumber(resource.getId(), connection);
        if (versionNumber == -1) {
            versionNumber = 1;
        } else {
            //incrementing the version number
            versionNumber = versionNumber + 1;
        }
        statement.setLong(2, versionNumber);

        Object content = resource.getContent();
        if (content != null && content instanceof byte[]) {
            statement.setBytes(3, (byte[])content);
        } else {
            statement.setBytes(3, null);
        }

        statement.setString(4, resource.getLastUpdaterUserName());
        statement.setTimestamp(5, new Timestamp(System.currentTimeMillis()));
        statement.executeUpdate();

        return versionNumber;
    }

    private void updateDependency(Resource resource,
                                  long versionNumber,
                                  boolean delete, Connection connection) throws SQLException , RegistryException {

        // updating dependency table , so that we can rollback to a particular version
        // updating for the current resource
        long parentNextVersionNumber =
                updateDependencyTable(resource, versionNumber, delete, connection);

        // a/b/c/r1.wsdl
        // Next a/b/c
        // Next a/b
        // Next a
        if (parentNextVersionNumber != -1) {

            // updating its parent
            String path = resource.getPath();
            while (path != null && path.indexOf("/") > -1) {

                int lastIndex = path.lastIndexOf("/");
                String parentPath = path.substring(0, lastIndex);
                Resource parentArtifact = getLatestVersion(parentPath, connection);

                if (parentArtifact != null) {
                    path = parentArtifact.getPath();
                    parentArtifact.setLastUpdaterUserName(resource.getLastUpdaterUserName());
                    parentNextVersionNumber = updateDependencyTable(
                            parentArtifact, parentNextVersionNumber, false, connection);

                    if (parentNextVersionNumber == -1) {
                        break;
                    }
                } else {
                    break;
                }
            }
        }
    }

    /**
     * @param resource      : Artifact we need to update itself and its immediate parent
     * @param versionNumber : Version number of the artifact
     * @param delete        : delete or adding
     * @param connection    : DB connection
     * @return : Will return the resource version of the parent resource
     * @throws SQLException : if something went wrong
     */
    private long updateDependencyTable(Resource resource,
                                       long versionNumber,
                                       boolean delete, Connection connection) throws SQLException , RegistryException {
        /**
         *        D0
         *    D1      D2
         *  D3  R1    R2
         *
         *  First fine D1 V0
         */
        // Checking whether the path is null or top most directory
        // Checking whether resource is root node or not

        String path = resource.getPath();

        // Checking whether the path is null or top most directory
        // if (path != null && path.indexOf("/") > -1) {
        if (path != null && !path.equals(RegistryConstants.ROOT_PATH)) {

            int lastIndex = path.lastIndexOf("/");
            String parentPath = RegistryConstants.ROOT_PATH;
            if (lastIndex != 0) {
                parentPath = path.substring(0, lastIndex);
            }

            if ("".equals(parentPath)) {
                //TODO : Need to think the logic here

            } else {

                long parentId = getResourceID(parentPath, connection);

                if (resource.getLastModified() == null) {
                    resource.setLastModified(new java.util.Date());
                }
                // Need to get the parent's parent and update the and update the

                long artifactversionNumber = getCurrentVerisonNumber(parentPath, connection);
//                long artifactversionNumber = getLatestVerisonNumber(parentId, connection);

                long nextVersionNumber = createNextVersion(parentId,
                                                           resource.getLastUpdaterUserName(),
                                                           resource.getLastModified().getTime(),
                                                           connection);

                String selectSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
                PreparedStatement s = connection.prepareStatement(selectSql);
                s.setLong(1, parentId);
                s.setLong(2, artifactversionNumber);
                ResultSet result = s.executeQuery();

                ArrayList dependeyMap = new ArrayList();

                // to see whether the parent node is there in the table , if not need to add that
                while (result.next()) {

                    long daid = result.getLong("DAID");

                    if (daid != resource.getId()) {
                        long dvn = result.getLong("DVN");
                        PreparedStatement ps = getStatementForDependencyTable(connection,
                                                                              parentId,
                                                                              nextVersionNumber,
                                                                              daid, dvn);
                        dependeyMap.add(ps);
                    }
                }

                if (!delete) {
                    PreparedStatement ds = getStatementForDependencyTable(connection,
                                                                          parentId,
                                                                          nextVersionNumber,
                                                                          resource.getId(),
                                                                          versionNumber);
                    dependeyMap.add(ds);
                }

                for (int i = 0; i < dependeyMap.size(); i++) {
                    PreparedStatement preparedStatement = (PreparedStatement)dependeyMap.get(i);
                    preparedStatement.executeUpdate();
                }

                return nextVersionNumber;
            }

        } else {
            // TODO , I think we do not need to add the entry here , since when adding a child node,
            // if the parent is not there then we will add that
        }

        return -1;
    }

    public long getLatestVerisonNumber(long resourceId, Connection connection) throws SQLException {

        String sql = "SELECT MAX(VN) FROM VERSIONS WHERE AID=?";
        PreparedStatement s = connection.prepareStatement(sql);
        s.setLong(1, resourceId);
        ResultSet result = s.executeQuery();

        if (result.next()) {
            return result.getLong(1);
        }

        return -1;
    }

    public String[] getVersionNumbers(long resourceId, Connection connection) throws SQLException {

        String SQL = "SELECT VN FROM VERSIONS WHERE AID=? ORDER BY UPDATED_TIME DESC";
        PreparedStatement s = connection.prepareStatement(SQL);
        s.setLong(1, resourceId);
        ResultSet result = s.executeQuery();
        ArrayList versionNumbers = new ArrayList();
        while (result.next()) {
            versionNumbers.add("" + result.getLong("VN"));
        }
        return (String[])versionNumbers.toArray(new String[versionNumbers.size()]);
    }

    public void restore(long resourceId, long toVersion, Connection connection)
            throws SQLException , RegistryException {

        Resource oldArtifact = getResourceByID(resourceId, toVersion, connection);
        if (User.getCurrentUser() != null) {
            oldArtifact.setLastUpdaterUserName(User.getCurrentUser());
        }
        String plainPath = oldArtifact.getPath();
        if (oldArtifact.getPath().indexOf("?v=") != -1) {
            String[] parts = oldArtifact.getPath().split("\\?v=");
            plainPath = parts[0];
        }
        update(plainPath, oldArtifact, connection);
        addResourceVersion(oldArtifact, connection);

        if (oldArtifact.isDirectory()) {
            activateDirectory(oldArtifact,connection);
            String selectChildSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
            PreparedStatement selectps = connection.prepareStatement(selectChildSql);
            selectps.setLong(1, resourceId);
            selectps.setLong(2, toVersion);
            ResultSet childResult = selectps.executeQuery();
            ArrayList dependencyList = new ArrayList();
            while (childResult.next()) {

                long daid = childResult.getLong("DAID");
                markActivated(daid,connection);
                long dvn = childResult.getLong("DVN");
                long versionNumber = getLatestVerisonNumber(resourceId, connection);
                PreparedStatement ps = getStatementForDependencyTable(connection,
                                                                      resourceId, versionNumber,
                                                                      daid, dvn);
                dependencyList.add(ps);
            }

            for (int i = 0; i < dependencyList.size(); i++) {
                PreparedStatement preparedStatement = (PreparedStatement)dependencyList.get(i);
                preparedStatement.executeUpdate();
            }
        }
    }

    /**
     * To delete a resource , what actually happen here is that update the version table with the new
     * data and resource will not actually removed from the table
     *
     * @param resource   : Resource to be deleted
     * @param connection : Connection to the DB
     * @throws java.sql.SQLException : if something went wrong
     */
    public void deleteResource(Resource resource,
                               Connection connection) throws SQLException  , RegistryException{
        long versionNumber = getLatestVerisonNumber(resource.getId(), connection);
        updateDependency(resource, versionNumber, true, connection);
    }

    private long createNextVersion(long resourceID,
                                   String lastUpdaterUserName,
                                   long lastModifiedTime,
                                   Connection connection) throws SQLException {

        long maxVersion = getLatestVerisonNumber(resourceID, connection);
        if (maxVersion == -1) {
            maxVersion = 0;
        }

        // here maxVersion is always >0
        maxVersion = maxVersion + 1;

        Resource resource = new Resource();
        resource.setId(resourceID);
        resource.setLastUpdaterUserName(lastUpdaterUserName);
        resource.setLastModified(new Date(lastModifiedTime));
        insertToVersionTable(resource, connection);

        return maxVersion;
    }

    private PreparedStatement getStatementForDependencyTable(Connection connection,
                                                             long parentId,
                                                             long nextVersionNumber,
                                                             long daid,
                                                             long dvn) throws SQLException {

        String SQL = "INSERT INTO DEPENDENCY (AID, VN , DAID , DVN) VALUES (?,? ,?,?)";
        PreparedStatement ps = connection.prepareStatement(SQL);
        ps.setLong(1, parentId);
        ps.setLong(2, nextVersionNumber);
        ps.setLong(3, daid);
        ps.setLong(4, dvn);
        return ps;
    }


    /**
     * This method is use to see whether a give resource is active or not , one can delete a resource
     * and try to access that giving the path of the resource in that case we should not return the
     * resource if it is deleted.
     *
     * @param resourcePath : Path of the resource
     * @param connection   : Connection to DB
     * @return : return true if the resource is active , else false
     */
    public boolean isResourceActive(String resourcePath, Connection connection) {
        String[] parts = resourcePath.split(RegistryConstants.PATH_SEPARATOR);
        String currentPath = "";
        int current = 0;
        if (parts.length > 0) {
            while (current < parts.length) {
                currentPath = currentPath + RegistryConstants.PATH_SEPARATOR + parts[current];
                if (!getResourceStatus(resourcePath, connection)) {
                    return false;
                }
                current++;
            }
        }
        return true;
    }

    /**
     * Will check whether the status of the given resource is active in the DB ., this method differ
     * from isResourceActive because this is actually call the DB Not the other
     *
     * @param resourcePath : path to resource
     * @param connection   : connection to DB
     * @return : If something went wrong
     */
    private boolean getResourceStatus(String resourcePath, Connection connection) {
        try {
            String SQL = "SELECT STATE FROM ARTIFACTS WHERE PATH=?";
            PreparedStatement stateStm = connection.prepareStatement(SQL);
            stateStm.setString(1, resourcePath);
            ResultSet results = stateStm.executeQuery();
            if (results.next()) {
                int state = results.getInt(1);
                if (RegistryConstants.ACTIVE_STATE == state) {
                    return true;
                }
            }
        } catch (SQLException e) {
            log.debug("Inside getResourceStatus trying to access the path: " + resourcePath +
                      " error message" + e.getMessage());
        }
        return false;
    }

    private String getParentPath(String path) {

        String[] parts = path.split(RegistryConstants.PATH_SEPARATOR);
        String parentPath = "";

        if (parts.length > 2) {
            for (int i = 1; i < parts.length - 1; i++) {
                parentPath = parentPath + "/" + parts[i];

            }
        } else {
            parentPath = "/";
        }

        return parentPath;
    }

    /**
     * This method will first split the given path and then try to create directory for each part
     * for example if the path is "/abc/xyx/r1" , then first it will see whether the resource "/abc"
     * is there if it is not there then it will create that else go to "/abc/xyz" , and see whether
     * the resource is there , if so continue the same procedure
     * @param path : Location where we need to add the resource
     * @param conn : DB connection
     * @param userID : user
     * @throws SQLException : Something went wrong
     * @throws RegistryException : Something went wrong
     * @param realm : Realm
     */
    private void createParentCollections(String path,
                                         Connection conn,
                                         String userID ,
                                         Realm realm) throws SQLException, RegistryException {
        String[] parts = path.split(RegistryConstants.PATH_SEPARATOR);

        String currentPath = RegistryConstants.ROOT_PATH + parts[1];
        for (int i = 2; i < parts.length; i++) {
            Resource currentCollection = getLatestVersion(currentPath, conn);
            if (currentCollection == null) {
                currentCollection = new Resource();
                currentCollection.setPath(currentPath);
                currentCollection.setDirectory(true);
                currentCollection.setAuthorUserName(userID);
                currentCollection.setLastUpdaterUserName(userID);
                add(currentCollection, conn);
                AuthorizationUtil.setDefaultAuthorizations(realm, currentPath, userID);
            }
            currentPath = currentPath + RegistryConstants.PATH_SEPARATOR + parts[i];
        }
    }

    private void deleteDirectory(Resource collection, Connection conn) throws SQLException , RegistryException {
        List childList = getChildren(collection.getId(), -1, conn);
        Iterator i = childList.iterator();
        while (i.hasNext()) {
            Resource child = (Resource)i.next();
            if (child.isDirectory()) {
                deleteDirectory(child, conn);
            } else {
                markDeleted(child.getPath(), conn);
                deleteResource(child, conn);
            }
        }
        markDeleted(collection.getPath(), conn);
        deleteResource(collection, conn);
    }

    private void activateDirectory(Resource collection, Connection conn) throws SQLException {
        List childList = getChildren(collection.getId(), -1, conn);
        Iterator i = childList.iterator();
        while (i.hasNext()) {
            Resource child = (Resource)i.next();
            if (child.isDirectory()) {
                activateDirectory(child, conn);
            } else {
                markActivated(child.getId(), conn);
            }
        }
        markActivated(collection.getId(), conn);
    }
  
}
